"math dialect, modeling functions in python's `math` stdlib"  # This file is generated by gen.py

from kirin.dialects.math.stmts import (
    cos as cos,
    erf as erf,
    exp as exp,
    pow as pow,
    sin as sin,
    tan as tan,
    ulp as ulp,
    acos as acos,
    asin as asin,
    atan as atan,
    ceil as ceil,
    cosh as cosh,
    erfc as erfc,
    fabs as fabs,
    fmod as fmod,
    log2 as log2,
    sinh as sinh,
    sqrt as sqrt,
    tanh as tanh,
    asinh as asinh,
    atan2 as atan2,
    atanh as atanh,
    expm1 as expm1,
    floor as floor,
    gamma as gamma,
    isinf as isinf,
    isnan as isnan,
    log1p as log1p,
    log10 as log10,
    trunc as trunc,
    lgamma as lgamma,
    degrees as degrees,
    radians as radians,
    copysign as copysign,
    isfinite as isfinite,
    remainder as remainder,
)
from kirin.dialects.math.interp import MathMethodTable as MathMethodTable
from kirin.dialects.math.dialect import dialect as dialect
